"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ConfigService = exports.Preset = void 0;
// eslint-disable-next-line @typescript-eslint/no-var-requires
const BootstrapUtils_1 = require("./BootstrapUtils");
const symbol_sdk_1 = require("symbol-sdk");
const CertificateService_1 = require("./CertificateService");
const LoggerFactory_1 = require("../logger/LoggerFactory");
const LogType_1 = require("../logger/LogType");
const NemgenService_1 = require("./NemgenService");
const model_1 = require("../model");
// eslint-disable-next-line @typescript-eslint/no-var-requires
const fs = require('fs');
var Preset;
(function (Preset) {
    Preset["bootstrap"] = "bootstrap";
    Preset["testnet"] = "testnet";
    Preset["light"] = "light";
})(Preset = exports.Preset || (exports.Preset = {}));
const logger = LoggerFactory_1.default.getLogger(LogType_1.LogType.System);
class ConfigService {
    constructor(root, params) {
        this.root = root;
        this.params = params;
    }
    toConfig(account) {
        return {
            privateKey: account.privateKey,
            publicKey: account.publicAccount.publicKey,
            address: account.address.plain(),
        };
    }
    static getNetworkIdentifier(networkType) {
        switch (networkType) {
            case symbol_sdk_1.NetworkType.MAIN_NET:
                return 'public';
            case symbol_sdk_1.NetworkType.TEST_NET:
                return 'public-test';
            case symbol_sdk_1.NetworkType.MIJIN:
                return 'mijin';
            case symbol_sdk_1.NetworkType.MIJIN_TEST:
                return 'mijin-test';
        }
        throw new Error(`Invalid Network Type ${networkType}`);
    }
    static getNetworkName(networkType) {
        switch (networkType) {
            case symbol_sdk_1.NetworkType.MAIN_NET:
                return 'public';
            case symbol_sdk_1.NetworkType.TEST_NET:
                return 'publicTest';
            case symbol_sdk_1.NetworkType.MIJIN:
                return 'mijin';
            case symbol_sdk_1.NetworkType.MIJIN_TEST:
                return 'mijinTest';
        }
        throw new Error(`Invalid Network Type ${networkType}`);
    }
    generateAddresses(networkType, size) {
        return ConfigService.getArray(size).map(() => this.toConfig(symbol_sdk_1.Account.generateNewAccount(networkType)));
    }
    async generateNodeAccount(index, node, networkType) {
        const type = node.type;
        const name = node.name || `${type}-${index}`;
        const signing = this.toConfig(symbol_sdk_1.Account.generateNewAccount(networkType));
        const vrf = this.toConfig(symbol_sdk_1.Account.generateNewAccount(networkType));
        const ssl = await new CertificateService_1.CertificateService(this.root, this.params).run(name);
        const friendlyName = node.friendlyName || ssl.publicKey.substr(0, 7);
        return { signing, vrf, ssl, type, name, friendlyName };
    }
    async generateNodeAccounts(networkType, nodes) {
        return Promise.all(nodes.map((node, index) => {
            return this.generateNodeAccount(index, node, networkType);
        }));
    }
    generateAddress(networkType) {
        return this.toConfig(symbol_sdk_1.Account.generateNewAccount(networkType));
    }
    static getArray(size) {
        return [...Array(size).keys()];
    }
    async run() {
        const configFolder = `${this.params.target}/config`;
        if (this.params.reset) {
            BootstrapUtils_1.BootstrapUtils.deleteFolder(configFolder);
        }
        if (fs.existsSync(configFolder)) {
            logger.info('Config folder exist, ignoring configuration. (run -r to reset)');
            const presetData = BootstrapUtils_1.BootstrapUtils.loadExistingPresetData(this.params.target);
            const addresses = BootstrapUtils_1.BootstrapUtils.loadExistingAddresses(this.params.target);
            return { presetData, addresses };
        }
        const presetData = BootstrapUtils_1.BootstrapUtils.loadPresetData(this.root, this.params.preset, this.params.assembly, this.params.customPreset);
        await fs.promises.mkdir(`${this.params.target}/config/generated-addresses`, { recursive: true });
        const networkType = presetData.networkType;
        const addresses = await this.generateRandomConfiguration(networkType, presetData);
        this.completePresetDataWithRandomConfiguration(presetData, addresses, networkType);
        await BootstrapUtils_1.BootstrapUtils.writeYaml(`${this.params.target}/config/generated-addresses/addresses.yml`, addresses);
        await this.generateNodes(presetData, addresses);
        await this.generateGateways(presetData, addresses);
        await BootstrapUtils_1.BootstrapUtils.mkdir(`${this.params.target}/data/nemesis-data/seed/00000`);
        if (presetData.nemesis) {
            await this.generateNemesisConfig(presetData, addresses);
        }
        else {
            const copyFrom = presetData.nemesisSeedFolder || `${this.root}/presets/${this.params.preset}/seed`;
            const copyTo = `${this.params.target}/data/nemesis-data/seed`;
            await BootstrapUtils_1.BootstrapUtils.generateConfiguration({}, copyFrom, copyTo);
        }
        await BootstrapUtils_1.BootstrapUtils.writeYaml(`${this.params.target}/config/preset.yml`, presetData);
        logger.info(`Configuration generated.`);
        return { presetData, addresses };
    }
    completePresetDataWithRandomConfiguration(presetData, addresses, networkType) {
        presetData.networkIdentifier = ConfigService.getNetworkIdentifier(presetData.networkType);
        presetData.networkName = ConfigService.getNetworkName(presetData.networkType);
        if (!presetData.nemesisGenerationHashSeed) {
            presetData.nemesisGenerationHashSeed = addresses.nemesisGenerationHashSeed;
        }
        //How can it work?
        const ownerAddress = symbol_sdk_1.Address.createFromPublicKey(presetData.nemesisSignerPublicKey, networkType);
        if (!presetData.currencyMosaicId)
            presetData.currencyMosaicId = BootstrapUtils_1.BootstrapUtils.toHex(symbol_sdk_1.MosaicId.createFromNonce(symbol_sdk_1.MosaicNonce.createFromNumber(0), ownerAddress).toHex());
        if (!presetData.harvestingMosaicId) {
            if (!presetData.nemesis) {
                throw new Error('nemesis must be defined!');
            }
            if (presetData.nemesis.mosaics && presetData.nemesis.mosaics.length > 1) {
                presetData.harvestingMosaicId = BootstrapUtils_1.BootstrapUtils.toHex(symbol_sdk_1.MosaicId.createFromNonce(symbol_sdk_1.MosaicNonce.createFromNumber(1), ownerAddress).toHex());
            }
            else {
                presetData.harvestingMosaicId = presetData.currencyMosaicId;
            }
        }
    }
    async generateRandomConfiguration(networkType, presetData) {
        const addresses = {
            networkType: networkType,
            nemesisGenerationHashSeed: presetData.nemesisGenerationHashSeed || symbol_sdk_1.Account.generateNewAccount(networkType).publicAccount.publicKey,
        };
        if (presetData.nodes) {
            addresses.nodes = await this.generateNodeAccounts(networkType, presetData.nodes);
        }
        if (presetData.gateways) {
            addresses.gateways = this.generateAddresses(networkType, presetData.gateways.length);
        }
        if (presetData.nemesis) {
            addresses.nemesisSigner = this.generateAddress(networkType);
            if (!presetData.nemesis.nemesisSignerPrivateKey && addresses.nemesisSigner) {
                presetData.nemesis.nemesisSignerPrivateKey = addresses.nemesisSigner.privateKey;
            }
            if (presetData.nemesis.mosaics) {
                const mosaics = {};
                presetData.nemesis.mosaics.forEach((m) => {
                    mosaics[m.name] = this.generateAddresses(networkType, m.accounts);
                });
                presetData.nemesis.mosaics.forEach((m) => {
                    const accounts = mosaics[m.name];
                    if (!m.currencyDistributions) {
                        m.currencyDistributions = accounts.map((a) => {
                            return { address: a.address, amount: m.supply / m.accounts };
                        });
                    }
                });
                addresses.mosaics = mosaics;
            }
        }
        if (!presetData.nemesisSignerPublicKey && addresses.nemesisSigner) {
            presetData.nemesisSignerPublicKey = addresses.nemesisSigner.publicKey;
        }
        return addresses;
    }
    async generateNodes(presetData, addresses) {
        await Promise.all((addresses.nodes || []).map(async (account, index) => {
            await this.generateNodeConfiguration(account, index, presetData, addresses);
        }));
    }
    async generateNodeConfiguration(account, index, presetData, addresses) {
        const copyFrom = `${this.root}/config/node`;
        const name = account.name;
        const outputFolder = `${this.params.target}/config/${name}`;
        const nodePreset = (presetData.nodes || [])[index];
        const generatedContext = {
            name: name,
            friendlyName: (nodePreset === null || nodePreset === void 0 ? void 0 : nodePreset.friendlyName) || account.friendlyName,
            harvesterSigningPrivateKey: account.signing.privateKey,
            harvesterVrfPrivateKey: account.vrf.privateKey,
        };
        const templateContext = Object.assign(Object.assign(Object.assign({}, presetData), generatedContext), nodePreset);
        await BootstrapUtils_1.BootstrapUtils.generateConfiguration(templateContext, copyFrom, outputFolder);
        await this.generateP2PFile(presetData, addresses, outputFolder, model_1.NodeType.PEER_NODE, 'peers-p2p.json');
        await this.generateP2PFile(presetData, addresses, outputFolder, model_1.NodeType.API_NODE, 'peers-api.json');
    }
    async generateP2PFile(presetData, addresses, outputFolder, type, jsonFileName) {
        var _a;
        const thisNetworkKnownPeers = (addresses.nodes || [])
            .map((node, index) => {
            if (node.type !== type) {
                return undefined;
            }
            const nodePresetData = (presetData.nodes || [])[index];
            const name = node.name;
            return {
                publicKey: node.ssl.publicKey,
                endpoint: {
                    host: name,
                    port: 7900,
                },
                metadata: {
                    name: name,
                    roles: nodePresetData.roles,
                },
            };
        })
            .filter((i) => i);
        const globalKnownPeers = ((_a = presetData.knownPeers) === null || _a === void 0 ? void 0 : _a[type]) || [];
        const data = {
            _info: `this file contains a list of ${type} peers`,
            knownPeers: [...thisNetworkKnownPeers, ...globalKnownPeers],
        };
        await fs.promises.writeFile(outputFolder + '/resources/' + jsonFileName, JSON.stringify(data, null, 2));
    }
    async generateNemesisConfig(presetData, addresses) {
        if (!presetData.nemesis) {
            throw new Error('nemesis must not be defined!');
        }
        const transactionDirectory = `${this.params.target}/config${presetData.nemesis.transactionsDirectory}`;
        await BootstrapUtils_1.BootstrapUtils.mkdir(transactionDirectory);
        const copyFrom = `${this.root}/config/nemesis`;
        const moveTo = `${this.params.target}/config/nemesis`;
        const templateContext = Object.assign(Object.assign({}, presetData), { addresses });
        await Promise.all((addresses.nodes || []).map((n) => this.createVrfTransaction(presetData, n)));
        //OPT IN!!
        // const namespacesBalances =
        //     storedTransactions
        //         .filter((t) => t.type === TransactionType.NAMESPACE_REGISTRATION)
        //         .map((t) => {
        //             const namespaceTransaction = t as NamespaceRegistrationTransaction;
        //             return parseInt(namespaceTransaction?.duration?.toString() || '0');
        //         })
        //         .reduce((p, n) => {
        //             return p + n;
        //         }, 0) || 0;
        if (presetData.nemesis.mosaics && (presetData.nemesis.transactions || presetData.nemesis.balances)) {
            logger.info('Opt In mode is ON!!! balances or transactions have been provided');
            if (presetData.nemesis.transactions) {
                const transactionHashes = [];
                const transactions = (await Promise.all(Object.entries(presetData.nemesis.transactions || {})
                    .map(([key, payload]) => {
                    const transactionHash = symbol_sdk_1.Transaction.createTransactionHash(payload, Array.from(symbol_sdk_1.Convert.hexToUint8(presetData.nemesisGenerationHashSeed)));
                    if (transactionHashes.indexOf(transactionHash) > -1) {
                        logger.warn(`Transaction ${key} wth hash ${transactionHash} already exist. Excluded from folder.`);
                        return undefined;
                    }
                    transactionHashes.push(transactionHash);
                    return this.storeTransaction(presetData, key, payload);
                })
                    .filter((p) => p))).filter((p) => p);
                logger.info(`Found ${transactions.length} opted in transactions.`);
            }
            const currencyMosaic = presetData.nemesis.mosaics[0];
            const nglAccount = currencyMosaic.currencyDistributions[0];
            const originalNglAccountBalance = nglAccount.amount;
            if (!nglAccount) {
                throw Error('"NGL" account could not be found for opt in!');
            }
            // if (namespacesBalances) {
            //     logger.info(`Removing namespace duration ${namespacesBalances} from "ngl" account ${namespacesBalances}`);
            // }
            let totalOptedInBalance = 0;
            if (presetData.nemesis.balances) {
                Object.entries(presetData.nemesis.balances || {}).forEach(([address, amount]) => {
                    totalOptedInBalance += amount;
                    currencyMosaic.currencyDistributions.push({ address, amount });
                });
                logger.info(`Removing ${Object.keys(presetData.nemesis.balances).length} accounts (total of ${totalOptedInBalance}) from "ngl" account ${nglAccount.address}`);
            }
            nglAccount.amount = nglAccount.amount - totalOptedInBalance;
            const providedBalances = Object.values(currencyMosaic.currencyDistributions)
                .map((d) => d.amount)
                .reduce((a, b) => a + b, 0);
            const currentBalance = providedBalances;
            // console.log('supply', currencyMosaic.supply);
            // console.log('providedBalances', providedBalances);
            // console.log('namespacesBalances', namespacesBalances);
            // console.log('currentBalance', currentBalance);
            if (nglAccount.amount < 1) {
                throw new Error(`NGL account didn't have enough balance (${originalNglAccountBalance}) to paid all the supplied optedin namespaces and accounts of ${currentBalance}`);
            }
            if (currentBalance !== currencyMosaic.supply) {
                throw new Error(`Current supplied balance of ${currentBalance} is different from expected supply of ${currencyMosaic.supply}`);
            }
        }
        await BootstrapUtils_1.BootstrapUtils.generateConfiguration(templateContext, copyFrom, moveTo);
        await new NemgenService_1.NemgenService(this.root, this.params).run(presetData);
    }
    async createVrfTransaction(presetData, node) {
        const deadline = symbol_sdk_1.Deadline['createFromDTO']('1');
        const vrf = symbol_sdk_1.VrfKeyLinkTransaction.create(deadline, node.vrf.publicKey, symbol_sdk_1.LinkAction.Link, presetData.networkType, symbol_sdk_1.UInt64.fromUint(0));
        const account = symbol_sdk_1.Account.createFromPrivateKey(node.signing.privateKey, presetData.networkType);
        const signedTransaction = account.sign(vrf, presetData.nemesisGenerationHashSeed);
        return await this.storeTransaction(presetData, `vrf_${node.name}`, signedTransaction.payload);
    }
    async storeTransaction(presetData, name, payload) {
        var _a;
        const transaction = symbol_sdk_1.TransactionMapping.createFromPayload(payload);
        const transactionsDirectory = `${this.params.target}/config${(_a = presetData.nemesis) === null || _a === void 0 ? void 0 : _a.transactionsDirectory}`;
        await fs.promises.writeFile(`${transactionsDirectory}/${name}.bin`, symbol_sdk_1.Convert.hexToUint8(payload));
        return transaction;
    }
    generateGateways(presetData, addresses) {
        return Promise.all((addresses.gateways || []).map((account, index) => {
            const copyFrom = `${this.root}/config/rest-gateway`;
            const generatedContext = {
                restPrivateKey: account.privateKey,
            };
            const gatewayPreset = (presetData.gateways || [])[index];
            const templateContext = Object.assign(Object.assign(Object.assign({}, presetData), generatedContext), gatewayPreset);
            const name = templateContext.name || `rest-gateway-${index}`;
            const moveTo = `${this.params.target}/config/${name}`;
            return BootstrapUtils_1.BootstrapUtils.generateConfiguration(templateContext, copyFrom, moveTo);
        }));
    }
}
exports.ConfigService = ConfigService;
ConfigService.defaultParams = { target: 'target', preset: Preset.bootstrap, reset: false };
